'use strict';

const Promise = require('bluebird');
const util = require('util');
const cloneDeep = require('clone-deep');
const utils = require('punch-utils');

const ALEXA_URL = (index, categories) => `http://www.alexa.com/topsites/category;${index}/Top/${categories.join('/')}`;
const BATCH_SIZE = 10;



const ScraperManager = require('../../index');
const scraperManager = new ScraperManager({
  scraperConfig: {
    strategy: {
      name: ScraperManager.STRATEGIES.HTTP
    },
    timeout: 8 * 1000,
    maxAttempts: 500,
    // Test to make sure result is an Alexa page and not a blocker.
    test: (result) => !!result.getHTML().match('Alexa - Top Sites by Category')
  },
  proxyManagerConfig: {
    proxyConfig: {
      // Logic for proxies to blacklist themselves.
      blacklistIf: (state) => {
        // Return if > 80% of the distinct URLs tested failed and at least
        // 15 have been attempted.
        return ((state.failed.distinctUrlCount/state.tried.distinctUrlCount) > .8) &&
            state.tried.distinctUrlCount > 15;
      }
    }
  }
});

scraperManager.start();


/**
 *
 */
function scrape(categoryChain, config) {
  config = config || {};
  categoryChain = Array.isArray(categoryChain) ? categoryChain : [categoryChain];

  const categoryUrls = categoryChain.map(formatCategoryForUrl);

  // Do initial scrape to get pagination count and categories.
  return scraperManager.scrape([ALEXA_URL(0, categoryUrls)])
  .then((results) => {
    const response = parseHTML(results[0]);

    // Generate all of the numbered urls based off of the maxPage.
    const urls = generateUrls(categoryUrls, response.maxPage);

    return scrapeListings(urls, config, categoryChain)
    .tap(listings => config.onListings && config.onListings(categoryChain, listings))
    .then((listings) => ({
      subcategories: response.subcategories,
      categoryChain,
      listings
    }));
  });
}


/**
 *
 */
function scrapeListings(urls, config, categoryChain) {
  config.queue  = config.queue || (new utils.promise.SlidingBatchQueue({
    batchSize: config.batchSize || BATCH_SIZE
  }));

  const listings = [];
  const index = new Set(); // Precautionary index to avoid adding duplicates.

  const listingScrapes = urls.map((url) => {
    return config.queue.resolve(() => scraperManager.scrape([url]))
    .then(results => {
      const result = parseHTML(results[0], true);
      result.listings.forEach(listing => {
        if (index.has(listing.url)) return;
        index.add(listing.url);
        listings.push(listing);
      });

      if (config.onListing) config.onListing(categoryChain, result.listings);
    });
  });

  return Promise.all(listingScrapes)
  .then(() => listings.sort((a, b) => a.ranking - b.ranking));
}


/**
 *
 */
function parseHTML(result, onlyListings) {
  const $ = result.getParsed();

  const formatted = {listings: parseListings($)};

  if (onlyListings) return formatted;
  return Object.assign(formatted, {
    maxPage: parseMaxPage($),
    subcategories: parseSubcategories($)
  });
}


/**
 *
 */
function parseMaxPage($) {
  const $paginationTiles = $('.pagination-page');
  let maxPage = 0;
  $paginationTiles.each(function() {
    let page = Number($(this).text().trim());
    if (isNaN(page)) return;
    if (page > maxPage) maxPage = page;
  });
  return maxPage;
}


/**
 *
 */
function parseSubcategories($) {
  const $subcategories = $('ul.subcategories a');
  const subcategories = [];

  $subcategories.each(function() {
    subcategories.push($(this).text().trim());
  });

  return subcategories;
};


/**
 *
 */
function parseListings($) {
  const $listings = $('.site-listing');
  const listings = [];

  $listings.each(function() {
    const formatted = {};
    const $listing = $(this);
    const $link = $listing.find('.desc-paragraph a');
    formatted.ranking = Number($listing.find('.count').text().trim());
    formatted.url = $link.text().trim().toLowerCase();
    formatted.alexa_url = `http://www.alexa.com${$link.attr('href').trim()}`;
    const $description = $listing.find('.description');
    formatted.description = $description.text().replace('…More', '').trim();
    listings.push(formatted);
  });

  return listings;
}


/**
 *
 */
function formatCategoryForUrl(category) {
  return category.replace(/\s/g, '_');
}


/**
 *
 */
function generateUrls(categories, cap) {
  cap = (cap + 1);
  // Handle weird case with alexa where you can get an extra page of links.
  if (cap == 21) cap = 22;
  let urls = [];
  // Not 0 indexed because first url has already been scraped.
  for (let i = 1; i < cap; i++) {
    urls.push(ALEXA_URL(i, categories));
  }
  return urls;
}


module.exports.scrape = scrape;
