'use strict';

const Promise = require('bluebird');
const uuid = require('uuid4');
const utils = require('punch-utils');

const categoryAndRankingScraper = require('./category-and-ranking-page-scraper');

const CATEGORY_BATCH_SIZE = 10;
const LISTING_BATCH_SIZE = 25;



/**
 * View readme for detailed usage instructions.
 * @param {
 *   onScrape: function,
 *
 *   fetchCategoriesFromDB: function, // Method to check DB for scraped data, in lieu of rescraping
 *   
 *   categoryQueue: utils.promise.SlidingBatchQueue, // Optional
 *   listingQueue: utils.promise.SlidingBatchQueue, // Optional
 * }
 */
function scrape(config) {
  config = config || {};
  config.categoryQueue = config.categoryQueue || (new utils.promise.SlidingBatchQueue({
    batchSize: config.batchSize || CATEGORY_BATCH_SIZE
  }));

  config.listingQueue = config.listingQueue || (new utils.promise.SlidingBatchQueue({
    batchSize: config.batchSize || LISTING_BATCH_SIZE
  }));

  config.fetchCategoriesFromDB = config.fetchCategoriesFromDB || (() => Promise.resolve([]));

  // Scrape Top Level categories.
  return categoryAndRankingScraper.scrape([])
  .then((result) => scrapeSubcategories([], result.subcategories, config));
}


/**
 * Recursively scrape subcategories. This promise is not resolved with any
 * meaningful value, simply an indication of its completion. Any action with
 * the results should be handled through the events as they occur, and there
 * is therefore no need to have a large result set kept in memory.
 */
function scrapeSubcategories(categoryChain, subcategories, config) {
  if (!subcategories || !subcategories.length) return Promise.resolve();

  const subcategoryChains = subcategories.map((subcategory) => {
    return categoryChain.concat(subcategory);
  });

  // Fetch any subcategory chains from DB. Any found will not be rescraped.
  return fetchCategoriesFromDB(subcategoryChains, config)
  // Scrape any category chains and their associated rank listings which were
  // not returned from the DB.
  .then(fromDBMap => {
    return subcategoryChains.map((subcategoryChain) => {
      const key = keyCategoryChain(subcategoryChain);
      if (fromDBMap[key]) return Promise.resolve(fromDBMap[key]);
      return scrapeCategoryChain(subcategoryChain, config);
    });
  })
  // Recurse to subcategories.
  .then(results => {
    return Promise.all(results.map((result) => {
      return scrapeSubcategories(result.categoryChain, result.subcategories, config);
    }));
  })
  .then(() => {}); // Enforce meaningless return.
}


/**
 *
 */
function fetchCategoriesFromDB(subcategoryChains, config) {
  return config.fetchCategoriesFromDB(subcategoryChains)
  .then((fromDB) => {
    return fromDB.reduce((map, result) => {
      map[keyCategoryChain(result.categoryChain)] = result;
      return map;
    }, {});
  });
}


/**
 *
 */
function scrapeCategoryChain(categoryChain, config) {
  console.log('Scraping category chain:', categoryChain);
  return config.categoryQueue.resolve(() => {
    return categoryAndRankingScraper.scrape(categoryChain, {
      queue: config.listingQueue,
      onListing: () => {
        console.log('Categories:', config.categoryQueue.getState());
        console.log('Listings:', config.listingQueue.getState());
      }
    });
  })
  .then(result => {
    if (config.onScrape) config.onScrape(result);
    return {
      subcategories: result.subcategories,
      categoryChain: result.categoryChain
    };
  });
}


/**
 *
 */
function keyCategoryChain(categoryChain) {
  return categoryChain.map(category => category.toLowerCase()).join('_');
}

module.exports.scrape = scrape;
