'use strict';

const deepFreeze = require('deep-freeze-strict');
const Promise = require('bluebird');
const DEFAULTS = require('./constants').DEFAULTS;
const ERROR = require('./constants').ERROR;
const AppErr = require('punch-error');
const utils = require('punch-utils');

const ProxyManager = require('punch-proxy-manager');

const scraper = require('./scraper/');
const ScrapeStatus = require('./scrape-status');

const BATCH_SIZE = 10;


class ScraperManager {
  /**
   *
   */
  constructor(config) {
    this.config = config = config || {};
    this.proxyManager = new ProxyManager(config.proxyManagerConfig || {});
    this.maxAttempts = config.maxAttempts || DEFAULTS.LINK_MAX_TRY;

    this.proxyManager.start();
  }
  
  start(){
    // Start loading proxies.
    return this.proxyManager.start();
  }

  stop(){
    return this.proxyManager.stop();
  }


  /**
   * @param Array<string> urls The list of urls to scrape.
   * @param {{
   *   batchSize: number, (the max amount of items to scrape at once)
   *   maxAttempts: number, (the max amount of attempts to make per url)
   *   onUpdate: function<updatedItemStatus, scrapeStatus>
   * }} config Config object.
   */
  scrape(urls, config) {
    config = Object.assign({}, this.config.scraperConfig || {}, config || {});
    let status = new ScrapeStatus({
      urls,
      onUpdate: config.onUpdate,
      maxAttempts: config.maxAttempts
    });

    config.batchSize = config.batchSize || BATCH_SIZE;

    return utils.promise.slidingBatch(urls, (url) => {
      return this.scrapeUrl(url, config, status);
    }, config.batchSize);
  }


  /**
   *
   */
  scrapeUrl(url, config, status) {
    config = config || {};
    const scraperConfig = Object.assign({},
      this.config.scraperConfig || {},
      config.scraperConfig || {});

    return this.proxyManager.ensureLoaded()
    .then(() => {
      const proxy = this.proxyManager.get();
      scraperConfig.proxy = proxy;
      proxy.markTried(url);

      return scraper.scrape(url, scraperConfig)
      // Test if html that is returned is valid.
      .tap(html => config.test ? runTest(config.test, html) : html)
      .tap((result) => {
        proxy.markSucceeded(url);
        status.update(url, proxy.getIP(), result);
      })
      .catch(err => {
        proxy.markFailed(url);
        status.update(url, proxy.getIP());
        if (!status.hasExceededAttempts(url)) return this.scrapeUrl(url, config, status);
        return AppErr.reject(err, ERROR.LINK_MAX_TRY_ERROR);
      });
    })
    .catch(err => {
      console.log('Error loading proxies:', err, err.stack);
      return this.scrapeUrl(url, config, status);
    });
  }
}

/**
 * Test if returned HTML should qualify as a successful scrape.
 */
function runTest(test, html) {
  return Promise.resolve(test(html))
  .tap(result => (result === false) && Promise.reject())
  .catch(err => {
    console.log('HTML returned did not pass test.');
    return Promise.reject(err);
  });
}

ScraperManager.STRATEGIES = require('./scraper/strategies/constants').STRATEGY;

module.exports = ScraperManager;
